/////////////////////////////////////////////////////////////////////////////
/*
  Copyright 2001 Ronald S. Burkey

  This file is part of GutenMark.

  GutenMark is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GutenMark is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GutenMark; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Filename:	AutoMark.c
  Purpose:	This is the main program for AutoMark, a vanilla-ANSI C 
  		program that marks up a vanilla-ASCII Project Gutenberg 
		(or similar) text file to HTML, with the intention of 
		making it printable on demand, or very beautifully readable.  
		In other words, Project Gutenberg has provided the content, 
		but now I want to make that content aesthetically pleasing.
  Mods:		08/30/01 RSB	Began.
  		09/03/01 RSB	Added MarkInsertChar.
		09/04/01 RSB	Some things that produced invalid HTML were 
				fixed:  '>', '<', '&' within body of the 
				text.  Empty <head></head> (no <title>).
				Empty anchor after Gutenberg header.  Open 
				paragraphs, headings, block quotes at end of 
				file.	
		09/06/01 RSB	Now allow output file as command-line 
				argument.  Also, allow some command-line 
				options (--no-justify, --help).  Added a lot 
				more processing on the "title".		
		09/08/01 RSB	Area between the PG header and the actual text, 
				which I call the "prefatory" area, is now now 
				detected and used essentially without markup.  
				Phrases found in this area (often a table of 
				contents) are used preferentially to detect 
				headings in the text and (conversely) to 
				reject identifications as headings.  MANY 
				additional tweaks performed.  Tables are now 
				handled, but only as "preformatted text" and 
				not a HTML tables.
		09/09/01 RSB	Again, numerous changes.
		09/15/01 RSB	Now handles PG so-called "8-bit ASCII".  A 
				lot more cleanup, such as block quotes, new 
				types of italic styles, and so on.
		09/16/01 RSB	Here is just a trivial remark added for 
				checking CVS.
		11/02/01 RSB	Added GPL disclaimer, and cleaned up the 
				formatting, prior to first web release.	
				Added startup messages.	
								
  Problems:	>	Greek phrases transliterated (without markup) into 
  			English.
		>	Footnotes embedded directly in the text, in braces 
			but without other markup, thusly:  
			"[text of footnote]"
		>	Footnotes in a "curly brace" style, collected at the 
			end of the text, thusly:  in the text body "{10}", 
			and at the end "{10} text of note".
		>	Replacement of "--" by something like &mdash; but I 
			find that '-', &mdash; and &ndash; all render as the 
			same narrow dash and not as a long dash.
		>	Should stuff like "to-"/"day" be  replaced 
			by "today" or by "to-day"?  The former requires
			building a spelling dictionary into the program.
			But in many circa-1900 documents, "to-day"
			and "to-morrow" are the actual word-forms
			expected (not just at the ends of lines).
		>	The preparer of the etext using in error
			a single dash without spaces to indicate a break
			in thought.  In other words, "-" where "--" or
			" -- " would be more appropriate.  There is no
					way of detecting this problem.
		>	There is no way to correctly recognize things like
			"right- and left-hand" (when "right-" appears at 
			the end of a line).  In other words, it would be
			converted to "right-and left-hand".  (This doesn't
			actually occur in the text, by the way.)
		>	Footnotes with markings in the body of the
			text like "[FN#10]", and with the footnotes 
			themselves collected at the end, like
			"[FN#10] text of footnote".
		>	Style in which italicized text is in all-caps.  Here 
			are some notes on the PROPOSED handling of these:
			1.  	In general, all-caps will be converted to 
				italicized all lower-case, except as indicated 
				below.
			2.	Words apparently at the beginnings of 
				sentences will have their first letter 
				capitalized.
			3.	Words appearing elswhere in the text NOT in 
				all-caps (and not at the beginnings of 
				sentences) will have the same capitalization 
				as in their other occurrences.
			4.	All-cap Roman Numbers will not be converted to 
				italics unless italicized text appears on one 
				or the other side of it as well.  "I" won't be 
				italicized unless surrounded by other 
				italicized text.  In either case, they will not 
				be converted to lower case.
			5.	Words from a known list of all-cap words (like 
				"USA", "HRH", "USS", and so on) will not be 
				converted to lower case and will not be 
				italicized unless surrounded by italicized 
				text.
			6.	Titles of books.  These will be recognized 
				by the presence of nearby non-italicized words 
				that are capitalized but not in all-caps -- 
				and thus might be the author's name.  In this 
				case, the phrase will be italicized, but will 
				be capitalized in the manner of titles.					
		>	Footnotes marked with "*", "**", etc., and collected 
			at the ends of chapters.
		>	Footnotes embedded in the text, but with numbers:  
			"[53]" in the text, and a line like "[53] Text of 
			footnote." after the end of the paragraph.
		>	Tabular data:  as in flymc10.txt.
		>	Endnote style marked in the text as (a), (b), (c), 
			etc.
		>	Alignment, as sometimes found in plays (or even in 
			verse, like 8faun10.txt) where the left half and 
			right half of the lines are separated by braces, 
			represented by {, |, (, or [ where multiple lines 
			are expected to line up.  For example:
						   ( Nonsense!
						   | You can't be serious!
			I've decided to leave you! { What!
						   | You haven't the nerve!
						   ( That's crazy talk!	
			In other words, the (, |, and { symbols are expected 
			to form a single large brace.
		>	Could use a better way to detect indented verse.
			Really should analyze the indentation patterns of 
			the stanzas, so that when something weird (but 
			indented properly) like a leading lower-case character 
			is encountered, it's treated properly.
  
  The program has to handle many special cases, such as differing paragraph 
  styles (indentation vs. no indentation, extra blank lines between paragraphs
  vs. no extra blank lines, etc.), differing styles for indicating italics 
  (or bold, or underlining), hyphenation vs. no hyphenation, Gutenberg vs. 
  non-Gutenberg, etc.  However, I want this program to be AUTOMATIC, so I'm 
  holding off on adding command-line switches for ANYTHING.  The program will 
  analyze the text to determine the styles, and if extra passes are necessary 
  to accomplish this, then so be it!
  
  The program works as follows:
  	AutoMark [options] [inputfile [outputfile]]
  If outputfile and/or inputfile are missing, then the stdout and/or stdin 
  are used.	
*/

#include <stdio.h>
#include "AutoMark.h"

void NormalizeTitle (char *s, int size);


//--------------------------------------------------------------------------

int
main (int argc, char *argv[])
{
  AnalysisDataset Dataset;
  int ErrorCode, c, Remove, AtEnd, i;
  MarkupRecord Mark;
  unsigned long Offset;
  char s[256];
  int FilesFound = 0;
  FILE *OutputFile;

  Dataset.IndentedParagraphs = 0;
  Dataset.TextStart = 0;

  // Process the command line.
  Dataset.InputFile = stdin;
  OutputFile = stdout;
  Dataset.NoJustify = 0;
  fprintf (stderr, "GutenMark PG-to-HTML markup tool, build " __DATE__ "\n");
  fprintf (stderr, "Copyright (C) 2001 Ronald Burkey (info@birdsproject.org)\n");
  fprintf (stderr, "GutenMark comes with ABSOLUTELY NO WARRANTY.\n");
  fprintf (stderr, "This is free software, licensed under the GNU GPL.\n");
  for (i = 1; i < argc; i++)
    {
      if (*argv[i] == '-')
	{
	  if (0 == strcmp (argv[i], "--help"))
	    {
	      printf ("----------------------------------------------------\n");
	      printf ("USAGE:");
	      printf
		("\tAutoMark [Options] [InputTxtFile [OutputHtmlFile]]\n");
	      printf
		("If the output file and/or input file are missing, "
		 " then the standard\n");
	      printf ("output and/or input are used.  "
		      "The available options are:\n");
	      printf ("\t--help           Displays this information.\n");
	      printf
		("\t--no-justify     Allows ragged right margin.  "
		 "Default is justified.\n");
	      return (0);
	    }
	  if (0 == strcmp (argv[i], "--no-justify"))
	    {
	      Dataset.NoJustify = 1;
	      continue;
	    }
	}
      else if (FilesFound == 0)
	{
	  Dataset.InputFile = fopen (argv[i], "rb");
	  if (Dataset.InputFile == NULL)
	    {
	      fprintf (stderr, "Input file does not exist.\n");
	      return (1);
	    }
	  FilesFound++;
	}
      else if (FilesFound == 1)
	{
	  OutputFile = fopen (argv[i], "w+b");
	  if (OutputFile == NULL)
	    {
	      fprintf (stderr, "Output file cannot be created.\n");
	      return (1);
	    }
	  FilesFound++;
	}
      else
	{
	  fprintf (stderr, "Too many filename listed on command line.\n");
	  fclose (Dataset.InputFile);
	  fclose (OutputFile);
	  return (1);
	}
    }


  Dataset.MarkupFile = tmpfile ();
  if (Dataset.MarkupFile == NULL)
    {
      fprintf (stderr, "Cannot create temporary markup file.\n");
      return (1);
    }

  // First stage:  Style analysis.  The analysis actually performs all 
  // of the markup, but doesn't modify the input or output files.  
  // Instead the  markups are saved to another file.
  ErrorCode = Markup (&Dataset);
  if (ErrorCode != 0)
    {
      fprintf (stderr, "An error code of %d occurred during markup.\n",
	       ErrorCode);
      return (1);
    }

  // Second stage:  Print output header.
  fprintf
    (OutputFile,
     "<!doctype html public \"-//w3c//dtd html 4.0 transitional//en\">\n");
  fprintf (OutputFile, "<html>\n");
  fprintf (OutputFile, "<head>\n");
  fseek (Dataset.InputFile, 0, SEEK_SET);
  fgets (s, sizeof (s) - 1, Dataset.InputFile);
  fgets (s + strlen (s), sizeof (s) - strlen (s) - 1, Dataset.InputFile);
  NormalizeTitle (s, sizeof (s));
  fprintf (OutputFile, "<TITLE>%s</TITLE>\n", s);
  fprintf (OutputFile, "</head>\n");
  fprintf (OutputFile, "<body>\n");
  fprintf (OutputFile, "<!-- Short-line cutoffs are %u and %u -->\n",
	   Dataset.ShortLineSize, Dataset.ReallyShortLineSize);

  // Third stage:  Print marked-up text.  This is basically a matter 
  // of merging the markup file with the input file.
  fseek (Dataset.MarkupFile, 0, SEEK_SET);
  fseek (Dataset.InputFile, 0, SEEK_SET);
  ErrorCode = fread (&Mark, sizeof (Mark), 1, Dataset.MarkupFile);
  if (ErrorCode != 1)
    Mark.Type = MarkNoMoreMarks;
  for (Offset = Remove = AtEnd = 0; (c = fgetc (Dataset.InputFile)) != EOF;
       Offset++)
    {
    LastShot:
      while (Offset == Mark.Offset && Mark.Type != MarkNoMoreMarks)
	{
	  switch (Mark.Type)
	    {
	    case MarkRemoveChar:
	      Remove = 1;
	      break;
	    case MarkInsertChar:
	      putc (Mark.Insert, OutputFile);
	      break;
	    case MarkNbsp:
	      fprintf (OutputFile, "&nbsp;");
	      break;
	    case MarkBeginJustifiedParagraph:
	      if (Dataset.NoJustify)
		fprintf (OutputFile, "<P>");
	      else
		fprintf (OutputFile, "<P align=justify>");
	      break;
	    case MarkBeginRaggedParagraph:
	      fprintf (OutputFile, "<P>");
	      break;
	    case MarkBeginCenteredParagraph:
	      fprintf (OutputFile, "<P align=center>");
	      break;
	    case MarkEndParagraph:
	      fprintf (OutputFile, "</P>");
	      break;
	    case MarkBeginItalics:
	      fprintf (OutputFile, "<I>");
	      break;
	    case MarkEndItalics:
	      fprintf (OutputFile, "</I>");
	      break;
	    case MarkBeginBold:
	      fprintf (OutputFile, "<B>");
	      break;
	    case MarkEndBold:
	      fprintf (OutputFile, "</B>");
	      break;
	    case MarkBeginUnderline:
	      fprintf (OutputFile, "<U>");
	      break;
	    case MarkEndUnderline:
	      fprintf (OutputFile, "</U>");
	      break;
	    case MarkBreak:
	      fprintf (OutputFile, "<BR>");
	      break;
	    case MarkBlockquote:	// I think, not used.
	      fprintf (OutputFile, "<BLOCKQUOTE>");
	      break;
	    case MarkEndBlockquote:	// I think, not used.
	      fprintf (OutputFile, "</BLOCKQUOTE>");
	      break;
	    case MarkHeader1:
	      fprintf (OutputFile, "<H1>");
	      break;
	    case MarkEndHeader1:
	      fprintf (OutputFile, "</H1>");
	      break;
	    case MarkBeginSubtitle:
	      fprintf (OutputFile, "<P><B>");
	      break;
	    case MarkEndSubtitle:
	      fprintf (OutputFile, "</B></P>");
	      break;
	    case MarkBeginTable:
	      fprintf (OutputFile, "<PRE>");
	      break;
	    case MarkEndTable:
	      fprintf (OutputFile, "</PRE>");
	      break;
	    case MarkJumpPastGutenbergHeader:
	      fprintf (OutputFile,
		       "<H1>Project Gutenberg Fine Print</H1><pre>\n");
	      break;
	    case MarkEndOfGutenbergHeader:
	      fprintf (OutputFile, "</pre>");
	      if (Dataset.LowestNonPrefatoryLine > 0)
		fprintf (OutputFile, "<H1>Prefatory Materials</H1>\n");
	      break;
	    }
	  ErrorCode = fread (&Mark, sizeof (Mark), 1, Dataset.MarkupFile);
	  if (ErrorCode != 1)
	    Mark.Type = MarkNoMoreMarks;
	}
      if (AtEnd)
	break;
      if (Remove)
	Remove = 0;
      else
	{
	  if (c == '<')
	    fprintf (OutputFile, "&lt;");
	  else if (c == '>')
	    fprintf (OutputFile, "&gt;");
	  else if (c == '&')
	    fprintf (OutputFile, "&amp;");
	  else if (c < 128)
	    putc (c, OutputFile);
	  else
	    fprintf (OutputFile, "&#%d;", c);
	}
    }
  // What this does is to allow us to process any remaining markups that
  // are supposed to occur at the very end of the input (such as </p>).  
  if (AtEnd == 0)
    {
      AtEnd = 1;
      goto LastShot;
    }

  // Last stage:  Close the output HTML stream.          
  fprintf (OutputFile, "</body>\n");
  fprintf (OutputFile, "</html>\n");
  fclose (Dataset.MarkupFile);
  fclose (Dataset.InputFile);
  fclose (OutputFile);
  return (0);

}

//---------------------------------------------------------------------
// The "title" has been read from the first line of the file.  
// However, it may have all kinds of Gutenbergy stuff, like asterisks, 
// the phrase "Project Gutenberg etext of", and so on.

void
NormalizeTitle (char *s, int size)
{
  char *ss;

  // Make sure the string is null-terminated.
  s[size - 1] = '\0';

  // Get rid of leading asterisks and white space.
  for (ss = s; *ss == '*' || isspace (*ss); ss++);
  if (ss > s)
    strcpy (s, ss);

  // Let's get rid of trailing asterisks and white space also.
  if (*s != '\0')
    {
      for (ss = s + strlen (s) - 1; *ss == '*' || isspace (*ss); ss--);
      ss[1] = '\0';
    }

  // Let's also get rid of multiple spaces.
  for (ss = s; *ss;)
    {
      if (isspace (*ss) && isspace (ss[1]))
	strcpy (ss, ss + 1);
      else if (*ss == '\t')
	*ss = ' ';
      else
	ss++;
    }

  // Look for some stuff that PG sometimes adds AFTER the title.
  for (ss = s; *ss; ss++)
    if (*ss == '*' || *ss == '#' || *ss == '(' || *ss == '[' || *ss == '{')
      {
	while (ss > s
	       && (ss[-1] == ' ' || ss[-1] == ',' || ss[-1] == '.'
		   || ss[-1] == ';' || ss[-1] == ':' || ss[-1] == '-'))
	  ss--;
	*ss = '\0';
	break;
      }

  // Now get rid of several common things that PG likes to add:
  if (0 == strncasecmp (s, "PG ", 3))
    strcpy (s, s + 3);
  if (0 == strncasecmp (s, "PG\'s ", 5))
    strcpy (s, s + 5);
  else if (0 == strncasecmp (s, "Project Gutenberg\'s ", 20))
    strcpy (s, s + 20);
  else if (0 == strncasecmp (s, "Project Gutenberg ", 18))
    strcpy (s, s + 18);
  else if (0 == strncasecmp (s, "The Project Gutenberg ", 22))
    strcpy (s, s + 22);
  if (0 == strncasecmp (s, "etext of ", 9))
    strcpy (s, s + 9);
  else if (0 == strncasecmp (s, "etext: ", 7))
    strcpy (s, s + 7);
  else if (0 == strncasecmp (s, "etext, ", 7))
    strcpy (s, s + 7);
  else if (0 == strncasecmp (s, "etext; ", 7))
    strcpy (s, s + 7);
  else if (0 == strncasecmp (s, "etext. ", 7))
    strcpy (s, s + 7);
  else if (0 == strncasecmp (s, "etext ", 6))
    strcpy (s, s + 6);

  // Finally, if the string is empty, fill in a default value.
  if (*s == '\0')
    strcpy (s, "A Project Gutenberg Book");

}
